/*
 *  drivers/mtd/nand/ts7350.c
 *
 *  Copyright (C) 2004 Technologic Systems (support@embeddedARM.com)
 *
 *  Derived from drivers/mtd/nand/edb7312.c
 *       Copyright (C) 2004 Marius Grger (mag@sysgo.de)
 *
 *  Derived from drivers/mtd/nand/autcpu12.c
 *       Copyright (c) 2001 Thomas Gleixner (gleixner@autronix.de)
 *
 * $Id: ts7350.c,v 1.2 2009/05/04 23:03:13 eddie Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   TS-7350 board which utilizes a Samsung 32 Mbyte part.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/sizes.h>

/*
 * MTD structure for TS7350 board
 */
static struct mtd_info *ts7350_mtd = NULL;

/* something is needed so nand_scan doesn't abort */
static unsigned char data_buf[1];
/* oob_buf is oob-per-page * pages-per-erase-block */
static unsigned char oob_buf[64*64];

/*
 * Module stuff
 */
static volatile unsigned int ts7350_nandctrl = 0;
static volatile unsigned int ts7350_nanddat = 0;

static struct mtd_partition partition_mbr[] = {
        { .name= "whole chip",
                  .offset= 0,
                  .size= 0x20000000 },
        { .name= "kernel",
                  .offset= 0x20000,
                  .size= 0x400000 },
        { .name= "initrd",
                  .offset= 0x420000,
                  .size= 0x400000 },
        { .name= "rootfs",
                  .offset= 0x820000,
                  .size= 0x20000000 - 0x820000 - 0x60000 },
        { .name= "part4",
                  .offset= 0x20000000 - 0x60000,
                  .size= 0x100000 },
};


/*
 *      hardware specific access to control-lines
 */
#define TS7350_NCE	(1<<2)
#define TS7350_CLE	(1<<1)
#define TS7350_ALE	(1<<0)
static void ts7350_hwcontrol(struct mtd_info *mtd, int cmd) 
{
	switch(cmd) {
		
	case NAND_CTL_SETCLE: 
		writeb(readb(ts7350_nandctrl) | TS7350_CLE, ts7350_nandctrl);
		break;
	case NAND_CTL_CLRCLE: 
		writeb(readb(ts7350_nandctrl) & ~TS7350_CLE, ts7350_nandctrl);
		break;
		
	case NAND_CTL_SETALE:
		writeb(readb(ts7350_nandctrl) | TS7350_ALE, ts7350_nandctrl);
		break;
	case NAND_CTL_CLRALE:
		writeb(readb(ts7350_nandctrl) & ~TS7350_ALE, ts7350_nandctrl);
		break;
		
	case NAND_CTL_SETNCE:
		writeb(readb(ts7350_nandctrl) | TS7350_NCE, ts7350_nandctrl);
		break;
	case NAND_CTL_CLRNCE:
		writeb(readb(ts7350_nandctrl) & ~TS7350_NCE, ts7350_nandctrl);
		break;
	}
}


/*
 *      read device ready pin
 */
static int ts7350_device_ready(struct mtd_info *mtd)
{
	return readb(ts7350_nandctrl) & 0x20;
}

static struct nand_oobinfo nand_oob_64 = {
	.useecc = MTD_NANDECC_AUTOPLACE,
	.eccbytes = 24,
	.eccpos = {
		40, 41, 42, 43, 44, 45, 46, 47, 
		48, 49, 50, 51, 52, 53, 54, 55, 
		56, 57, 58, 59, 60, 61, 62, 63},
	.oobfree = { {2, 38} }
};

/*
 * Main initialization routine
 */
static int __init ts7350_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	char *nandaddr;
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts = 0;

	/* Allocate memory for MTD device structure and private data */
	ts7350_mtd = kmalloc(sizeof(struct mtd_info) + 
			     sizeof(struct nand_chip),
			     GFP_KERNEL);
	if (!ts7350_mtd) {
		printk("Unable to allocate TS7350 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	
	/* Get pointer to private data */
	this = (struct nand_chip *) (&ts7350_mtd[1]);
	
	/* Initialize structures */
	memset((char *) ts7350_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));

	memcpy((char *)&ts7350_mtd->oobinfo,(char *)&nand_oob_64,sizeof(struct nand_oobinfo));

	/* Link the private data with the MTD structure */
	ts7350_mtd->priv = this;
	
 	nandaddr = __ioremap(0x600ff000, 0x1000, 0);

        /* insert callbacks */
        this->IO_ADDR_R = nandaddr + 0x406;
        this->IO_ADDR_W = nandaddr + 0x404;
        ts7350_nanddat = (unsigned int)nandaddr + 0x404;
        ts7350_nandctrl = (unsigned int)nandaddr + 0x400;

        this->hwcontrol = ts7350_hwcontrol;
        this->dev_ready = ts7350_device_ready;
        this->chip_delay = 15;

	this->eccmode = NAND_ECC_SOFT;
	ts7350_mtd->oobblock = 2048;
	ts7350_mtd->oobsize = 64;
	ts7350_mtd->eccsize = 256;
	
	this->options = 0;
	this->data_buf = data_buf;
	this->oob_buf = oob_buf;

	printk("Searching for NAND flash...\n");
printk("oobsize=0x%x\n", ts7350_mtd->oobsize);
	/* Scan to find existence of the device */
	if (nand_scan (ts7350_mtd, 1)) {
		kfree (ts7350_mtd);
		return -ENXIO;
	}

	//EWD: hardcoded partition table for TS7350
	mtd_parts = partition_mbr;
	mtd_parts_nb = 4;
	part_type = "static";
	
	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(ts7350_mtd, mtd_parts, mtd_parts_nb);
	
	/* Return happy */
	return 0;
}
module_init(ts7350_init);

/*
 * Clean up routine
 */
static void __exit ts7350_cleanup (void)
{
  //	struct nand_chip *this = (struct nand_chip *) &ts7350_mtd[1];
	
	/* Unregister the device */
	del_mtd_partitions(ts7350_mtd);
	del_mtd_device (ts7350_mtd);
	
	/* Free the MTD device structure */
	kfree (ts7350_mtd);
}
module_exit(ts7350_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jesse Off <joff@embeddedARM.com>");
MODULE_DESCRIPTION("MTD map driver for Technologic Systems TS-7350 board");
