/*
 *  drivers/mtd/nand/ts7400.c
 *
 *  Copyright (C) 2004 Technologic Systems (support@embeddedARM.com)
 *
 *  Derived from drivers/mtd/nand/edb7312.c
 *       Copyright (C) 2004 Marius Grger (mag@sysgo.de)
 *
 *  Derived from drivers/mtd/nand/autcpu12.c
 *       Copyright (c) 2001 Thomas Gleixner (gleixner@autronix.de)
 *
 * $Id: ts7400.c,v 1.1 2008/06/06 22:09:15 joff Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   TS7400 board which utilizes a Samsung 32 Mbyte part.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/sizes.h>

struct partition_info { // little-endian format
  unsigned int bootable:8; // 0x80 = active, 0x00 = don't use for booting
  unsigned int starting_head:8;
  unsigned int starting_csect:8;
  unsigned int starting_xcyl:8;
  // MSB bits 0-5 are sector
  unsigned int system_id:8; // format of partition
  unsigned int ending_head:8;
  unsigned int ending_csect:8;
  unsigned int ending_xcyl:8;
  unsigned int relative_sector:32; // sector offset from start of disk to start of volume
  unsigned int total_sectors:32;
} __attribute__((packed));

struct partition_table {
  struct partition_info part[4];
} __attribute__((__packed__));

struct MBR {
  char bootcode[446];
  struct partition_table pt;
  unsigned int sig55:8;
  unsigned int sigAA:8;
} __attribute__((__packed__));

/*
 * MTD structure for TS7400 board
 */
static struct mtd_info *ts7400_mtd = NULL;
/* something is needed so nand_scan doesn't abort */
static unsigned char data_buf[1];
/* oob_buf is oob-per-page * pages-per-erase-block */
static unsigned char oob_buf[64*64];

static struct nand_oobinfo nand_oob_64 = {
	.useecc = MTD_NANDECC_AUTOPLACE,
	.eccbytes = 24,
	.eccpos = {
		40, 41, 42, 43, 44, 45, 46, 47, 
		48, 49, 50, 51, 52, 53, 54, 55, 
		56, 57, 58, 59, 60, 61, 62, 63},
	.oobfree = { {2, 38} }
};


/*
 * Module stuff
 */

static unsigned int ts7400_nandctrl = IO_ADDRESS(TS7XXX_NAND_FLASH_CTRL);
static unsigned int ts7400_nandbusy = IO_ADDRESS(TS7XXX_NAND_FLASH_BUSY);

#ifdef CONFIG_MTD_PARTITIONS
/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info32[] = {
        { .name= "TS-BOOTROM",
                  .offset= 0,
                  .size= 0x4000 },
        { .name= "Linux",
                  .offset= 0x4000,
                  .size= 0x1d00000 },
        { .name= "RedBoot",
                  .offset= 0x1d04000,
                  .size= 0x2fc000 },
};
/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info128[] = {
        { .name= "TS-BOOTROM",
                  .offset= 0,
                  .size= 0x4000 },
        { .name= "Linux",
                  .offset= 0x4000,
                  .size= 0x7d00000 },
        { .name= "RedBoot",
                  .offset= 0x7d04000,
                  .size= 0x2fc000 },
};

static struct mtd_partition partition_mbr[] = {
	{ .name= "whole chip",
		  .offset= 0,
		  .size= 0x8000000 },
	{ .name= "kernel",
		  .offset= 0x20000,
		  .size= 0x200000 },
	{ .name= "initrd",
		  .offset= 0x220000,
		  .size= 0x200000 },
	{ .name= "rootfs",
		  .offset= 0x440000,
		  .size= 0x8000000 - 0x440000 },
	{ .name= "part4",
		  .offset= 0x2000000,
		  .size= 0 },
};

extern int parse_cmdline_partitions(struct mtd_info *master, 
				    struct mtd_partition **pparts,
				    const char *mtd_id);
#endif


static void ts7400_hwcontrol(struct mtd_info *mtd, int cmd) 
{
	switch(cmd) {
		
	case NAND_CTL_SETCLE: 
		writeb(readb(ts7400_nandctrl) | 0x2, ts7400_nandctrl);
		break;
	case NAND_CTL_CLRCLE: 
		writeb(readb(ts7400_nandctrl) & ~0x2, ts7400_nandctrl);
		break;
		
	case NAND_CTL_SETALE:
		writeb(readb(ts7400_nandctrl) | 0x1, ts7400_nandctrl);
		break;
	case NAND_CTL_CLRALE:
		writeb(readb(ts7400_nandctrl) & ~0x1, ts7400_nandctrl);
		break;
		
	case NAND_CTL_SETNCE:
		writeb(readb(ts7400_nandctrl) | 0x4, ts7400_nandctrl);
		break;
	case NAND_CTL_CLRNCE:
		writeb(readb(ts7400_nandctrl) & ~0x4, ts7400_nandctrl);
		break;
	}
}

/*
 *	read device ready pin
 */
static int ts7400_device_ready(struct mtd_info *mtd)
{
	return readb(ts7400_nandbusy) & 0x20;
}

static void ts7400_enable_hwecc(struct mtd_info *mtd,int dir) {
  volatile int *ts7x = (void *)IO_ADDRESS(TS7XXX_NAND_FLASH_CTRL);
  int dummy;

  dummy = *ts7x;
  //printk("enable_hwecc\n");
}

static int ts7400_calculate_ecc(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code) {
     volatile int *ts7x = (void *)IO_ADDRESS(TS7XXX_NAND_FLASH_CTRL);
     int dummy;

     dummy =  0xFFFFFF & ((*ts7x >> 6) | 3);
     //printk("Hardware ECC: %06X ",dummy);
     ecc_code[0] = (dummy >> 16) & 0xFF;
     ecc_code[1] = (dummy >> 8) & 0xFF;
     ecc_code[2] = (dummy >> 0) & 0xFF;
     //printk("(%02X %02X %02X)\n",ecc_code[0],ecc_code[1],ecc_code[2]);
     return dummy;
}

static int detect_hwecc(struct nand_chip *this) {
  volatile int *ts7x = (void *)IO_ADDRESS(TS7XXX_NAND_FLASH_CTRL);
  int dummy;

  dummy = *ts7x;
  dummy = *ts7x;
  if ((dummy >> 8) != -1) {
    return 0;
  }
  writeb(0x1, this->IO_ADDR_W);
  dummy = *ts7x;
   return ((dummy >> 8) != -1);
}

extern int nand_correct_data(struct mtd_info *mtd, u_char *dat, u_char *read_ecc, u_char *calc_ecc);
/*
 * Main initialization routine
 */
static int __init ts7400_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts = 0;
	char buf[512];
	int i,got = 0,parts=0,hwecc = 0;
	unsigned ofs=0;
	volatile unsigned short *tsmodel;
	struct MBR *mbr = (struct MBR *)buf;

	/* Allocate memory for MTD device structure and private data */
	ts7400_mtd = kmalloc(sizeof(struct mtd_info) + 
			     sizeof(struct nand_chip),
			     GFP_KERNEL);
	if (!ts7400_mtd) {
		printk("Unable to allocate TS7400 NAND MTD device structure.\n");
		return -ENOMEM;
	}
	
	/* Get pointer to private data */
	this = (struct nand_chip *) (&ts7400_mtd[1]);
	
	/* Initialize structures */
	memset((char *) ts7400_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));
	memcpy((char *)&ts7400_mtd->oobinfo,(char *)&nand_oob_64,sizeof(struct nand_oobinfo));
	
	/* Link the private data with the MTD structure */
	ts7400_mtd->priv = this;
	
	
	/* insert callbacks */
	this->IO_ADDR_R = (void *)IO_ADDRESS(TS7XXX_NAND_FLASH_DATA);
	this->IO_ADDR_W = (void *)IO_ADDRESS(TS7XXX_NAND_FLASH_DATA);
	ts7400_nandbusy = IO_ADDRESS(TS7XXX_NAND_FLASH_BUSY);
	ts7400_nandctrl = IO_ADDRESS(TS7XXX_NAND_FLASH_CTRL);

	this->hwcontrol = ts7400_hwcontrol;
	this->dev_ready = ts7400_device_ready;
	this->chip_delay = 15;
	this->eccmode = NAND_ECC_SOFT;

	ts7400_mtd->oobblock = 2048;
	ts7400_mtd->oobsize = 64;
	ts7400_mtd->eccsize = 256;

	this->options = 0;
	this->data_buf = data_buf;
	this->oob_buf = oob_buf;

	printk("Searching for NAND flash...\n");
	/* Scan to find existence of the device */
	if (nand_scan (ts7400_mtd, 1)) {
		kfree (ts7400_mtd);
		return -ENXIO;
	}
	for (i=0;i<512;i++) {
	  buf[i] = 0;
	}

{

	  if (mbr->sigAA == 0xAA && mbr->sig55 == 0x55) {
	    for (i=0;i<4;i++) {
	      if (mbr->pt.part[i].total_sectors) {
		parts++;
		partition_mbr[i+1].offset = 
		  512 * mbr->pt.part[i].relative_sector;
		if (i < 2) {
		  partition_mbr[i+1].size = 
		    512 * mbr->pt.part[i+1].relative_sector -
		    512 * mbr->pt.part[i].relative_sector;
		} else {
		  partition_mbr[i+1].size = 
		    512 * mbr->pt.part[i].total_sectors;
		}
		ofs = partition_mbr[i+1].offset 
		  + partition_mbr[i+1].size;
	      }
	    }
	  } else {
	    parts = 4;
	  }
	  mtd_parts = partition_mbr;
	}
	if (mtd_parts_nb == 0)
	{
	  //mtd_parts = partition_info32;
		/*
		if (ts7400_mtd->size >= (128 * 0x100000))
			mtd_parts = partition_info128;
		*/
		mtd_parts_nb = parts;
		if (parts > 0) {
		  part_type = "MBR";
		} else {
		  part_type = "static";
		}
	}
	
	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(ts7400_mtd, mtd_parts, mtd_parts_nb);
	
	/* Return happy */
	return 0;
}
module_init(ts7400_init);

/*
 * Clean up routine
 */
static void __exit ts7400_cleanup (void)
{
  //	struct nand_chip *this = (struct nand_chip *) &ts7400_mtd[1];
	
	/* Unregister the device */
	del_mtd_partitions(ts7400_mtd);
	del_mtd_device (ts7400_mtd);
	
	/* Free the MTD device structure */
	kfree (ts7400_mtd);
}
module_exit(ts7400_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jesse Off <joff@embeddedARM.com>");
MODULE_DESCRIPTION("MTD map driver for Technologic Systems TS7400 board");
