/*
 * Ethernet driver for TS-7300 Altera Ethernet Core
 * 
 *  (c) 2009 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2009-06-03:mos:initial version
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/init.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/bitops.h>

#include <linux/mii.h>
#include <asm-arm/arch-ep93xx/regmap.h>

#define ETH_BASE_ADD 			0x72100000
#define AETH_REG_OFFSET			0x2000					
#define IRQ_ETH_0 			40
#define AETH_TOTAL_SIZE                 0x800
static const char* cardname = "ts7300-aeth";
#define EREG(x) (x/4)
#define __clear_user(add,len) memset((add),0,(len))

struct aeth_private {
  unsigned *eth;
  int in_tx;
  struct net_device_stats stats;
};

static void aeth_init(volatile unsigned *eth, unsigned char *mymac) {
  unsigned mac1,mac2;

  mac1 = (unsigned)mymac[0] + (((unsigned)mymac[1]) << 8) + (((unsigned)mymac[2]) << 16)+ (((unsigned)mymac[3]) << 24);
  mac2 = (unsigned)mymac[4] + (((unsigned)mymac[5]) << 8);
  eth[EREG(0x0C)] = mac1; //0x04030201;
  eth[EREG(0x10)] = mac2; //0x0605;

  eth[EREG(0x1C)] = 0x3F0;
  eth[EREG(0x20)] = 0x10;
  eth[EREG(0x24)] = 0x3F0;
  eth[EREG(0x28)] = 0x0;
  eth[EREG(0x2C)] = 0x8;
  eth[EREG(0x30)] = 0x8;
  eth[EREG(0x34)] = 0x8;
  eth[EREG(0x38)] = 0x3;
  eth[EREG(0x40)] = 0x2;
  eth[EREG(0x3C)] = 0x1;
  eth[EREG(0x5C)] = 0xC;
}

static inline int eth_rx_empty(volatile unsigned *eth) {
  return ((eth[EREG(0x504)]) & (1<<14)) != 0;
}

static inline int eth_tx_intr(volatile unsigned *eth) {
  return ((eth[EREG(0x504)]) & 1) != 0;
}

static inline int eth_tx_intrack(volatile unsigned *eth) {
  return eth[EREG(0x504)] |= 1;
}

static inline int eth_rx_sop(volatile unsigned *eth) {
  return ((eth[EREG(0x504)]) & (1<<20)) != 0;
}

static inline int eth_rx_eop(volatile unsigned *eth) {
  return ((eth[EREG(0x504)]) & (1<<19)) != 0;
}

static inline int eth_tx_full(volatile unsigned *eth) {
  return ((eth[EREG(0x504)]) & (1<<7)) != 0;
}

static inline int eth_rx_mod(volatile unsigned *eth) {
  return ((eth[EREG(0x504)]) >> 23) & 3;
}

static inline int eth_rx_err(volatile unsigned *eth) {
  return (((eth[EREG(0x504)]) >> 9) & 31);
}

static inline void eth_multihash(volatile unsigned *eth,int n,int val) {
  eth[EREG(0x100)+n] = val;
}

static inline void eth_tx_en(volatile unsigned *eth,int val) {
  if (val) {
    eth[EREG(0x08)] |= 1;
  } else {
    eth[EREG(0x08)] &= ~1;
  }
}

static inline void eth_rx_en(volatile unsigned *eth,int val) {
  if (val) {
    eth[EREG(0x08)] |= 2;
  } else {
    eth[EREG(0x08)] &= ~2;
  }
}

static inline void eth_reset(volatile unsigned *eth) {
  eth[EREG(0x08)] = 0x2000;
  while (eth[EREG(0x08)] & 0x2000);
}

static inline void eth_promisc_en(volatile unsigned *eth,int val) {
  if (val) {
    eth[EREG(0x08)] |= (1<<4);
  } else {
    eth[EREG(0x08)] &= ~(1<<4);
  }
}


static inline void eth_mymac_en(volatile unsigned *eth,int val) {
  if (val) {
    eth[EREG(0x08)] |= (1<<9);
  } else {
    eth[EREG(0x08)] &= ~(1<<9);
  }
}

static inline void eth_loop_en(volatile unsigned *eth,int val) {
  if (val) {
    eth[EREG(0x08)] |= (1<<15);
  } else {
    eth[EREG(0x08)] &= ~(1<<15);
  }
}

static inline unsigned rx_word(volatile unsigned *eth) {
  return eth[EREG(0x400)];
}

static inline unsigned rx_len(volatile unsigned *eth) {
  return eth[EREG(0x500)];
}

static inline void txq_word(volatile unsigned *eth,unsigned val) {
  eth[EREG(0x404)] = val;
}

static inline void tx_send(volatile unsigned *eth,int mod) {
  eth[EREG(0x500)] = mod;
}

//----------------------------------------------------------------------
static int aeth_open(struct net_device *dev);
static int aeth_start_xmit(struct sk_buff *skb, struct net_device *dev);
static int aeth_close(struct net_device *dev);
static void aeth_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static struct net_device_stats *aeth_get_stats(struct net_device *dev);
static void aeth_set_multicast_list(struct net_device *dev);

static int aeth_open(struct net_device *dev) {
	struct aeth_private *cep = (struct aeth_private *)dev->priv;
	int ret;
	/* Install our interrupt handler.
	 */
	ret = request_irq(IRQ_ETH_0, aeth_interrupt, SA_SHIRQ, cardname, (void *)dev);
	if (0 != ret) {
		printk("Unable to allocate IRQ %d\n",IRQ_ETH_0);
		return ret;
        }
	// Enable receiver and transmiter 
	eth_reset(cep->eth);
	eth_tx_en(cep->eth,1);
	eth_rx_en(cep->eth,1);
	eth_mymac_en(cep->eth,1);

	return 0;
}

static int aeth_close(struct net_device *dev) {
	struct aeth_private *cep = (struct aeth_private *)dev->priv;
	// Free interrupt hadler 
	free_irq(IRQ_ETH_0, (void *)dev);

	// Disable receiver and transmitesr 
	eth_reset(cep->eth);
	return 0;
}

static int aeth_start_xmit(struct sk_buff *skb, struct net_device *dev) {
  int i,len=skb->len,mod = skb->len % 4;
  //unsigned *word = (unsigned *)skb->data;
  unsigned char *byte = (unsigned char *)skb->data;
  volatile struct aeth_private *cep;
  volatile unsigned *q;

  //printk("%p ---\n",skb->data);
  cep = (struct aeth_private *)dev->priv;
  q = &cep->eth[EREG(0x404)];
  /*
  if (cep->in_tx == 1) {
    i=0;
    cli();
    while (!eth_tx_intr(cep->eth) && i<20) i++;
    if (i==20) {
      printk("DROP\n");
    } else {
      //netif_wake_queue(dev);
      eth_tx_intrack(cep->eth);
    }
    sti();  
  }
  */
  netif_stop_queue(dev);

  //printk("(%d)",skb->len);
  len >>= 2;
  if (mod > 0) len++;
  if (mod == 1 || mod == 3) mod++; // hack
  for (i=0;i<len;i++) {    
    //    txq_word(eth,*(word++));
    //*q = *(word++);
    *q = *(byte) + (*(byte+1)<<8) + (*(byte+2)<<16) + (*(byte+3)<<24);
    byte += 4;
  }
  if (eth_tx_full(cep->eth)) {
    cep->stats.tx_errors++;
  }
  tx_send(cep->eth,0);
  cep->stats.tx_packets++;
  /*
    cli();
    while (!eth_tx_intr(cep->eth) && i<20) i++;
    if (i==20) {
      printk("Zzzz\n");
    } else {
      netif_wake_queue(dev);
      eth_tx_intrack(cep->eth);
    }
    sti();  
  */
  cep->in_tx = 1;  
  dev_kfree_skb(skb);
  return 0;
}

static void aeth_rx(struct net_device *dev,volatile struct aeth_private *cep) {
  struct sk_buff *skb;
  int i,len,off=0,err;
  unsigned *word;

  skb = dev_alloc_skb(2000); // larger than maximum we expect...
  if (skb == NULL) {
    printk("%s: Memory squeeze, dropping packet.\n", dev->name);
    cep->stats.rx_dropped++;
  } else {
    skb_reserve(skb, 4);
    skb->dev = dev;
    word = (unsigned *)skb->tail;
    do {
      for (i=0;i<16;i++) {
	*(word++) = rx_word(cep->eth);
      }
      off += 16;
    } while (!eth_rx_eop(cep->eth) && off < 512);  
    len = rx_len(cep->eth) - eth_rx_mod(cep->eth);
    skb_put(skb,len);
    if ((err = eth_rx_err(cep->eth))) {
      //printk("rx err (%d)\n",len); // DEBUG
      kfree_skb(skb); //kfree_skb_fast?
      cep->stats.rx_errors++;
      if (err & 1) cep->stats.rx_length_errors++;
      if (err & 2) cep->stats.rx_crc_errors++;
      if (err & 4) cep->stats.rx_fifo_errors++;
      if (err & 8) cep->stats.rx_frame_errors++; //gmii phy err
      if (err & 16) cep->stats.rx_missed_errors++; //collision
    } else {
      //printk("}%d{",len); // DEBUG
      skb->protocol = eth_type_trans(skb,dev);
      netif_rx(skb);
      cep->stats.rx_packets++;
    }
  }
}

static void aeth_tx(struct net_device *dev,volatile struct aeth_private *cep) {
  netif_wake_queue(dev);
  //printk("tx int\n");
  eth_tx_intrack(cep->eth);
  cep->in_tx = 0;
}

static void aeth_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
  struct	net_device *dev = dev_id;
  volatile struct	aeth_private *cep;

  cep = (struct aeth_private *)dev->priv;
  while (!eth_rx_empty(cep->eth)) {
    aeth_rx(dev,cep);
  }
  if (eth_tx_intr(cep->eth)) {
    aeth_tx(dev,cep);
  }
  return;
}

static inline int byte_xor_hash(__u8 byte) {
  int bit= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  byte >>= 1;
  bit ^= byte & 1;
  return bit;
}

static inline int mac_hash(__u8 *mac) {
  return 
    (byte_xor_hash(mac[0]) << 0) |
    (byte_xor_hash(mac[1]) << 1) |
    (byte_xor_hash(mac[2]) << 2) |
    (byte_xor_hash(mac[3]) << 3) |
    (byte_xor_hash(mac[4]) << 4) |
    (byte_xor_hash(mac[5]) << 5);
}

static struct net_device_stats *aeth_get_stats(struct net_device *dev)
{
  struct aeth_private *cep = (struct aeth_private *)dev->priv;
  return &cep->stats;
}

static void aeth_set_multicast_list(struct net_device *dev) {
  struct aeth_private *cep;  
  struct dev_mc_list *mc_ptr;
  int i,hash[64];

  cep = (struct aeth_private *)dev->priv;

  if (dev->flags & IFF_PROMISC) {
    eth_promisc_en(cep->eth,1);
    return;
  } else {
    eth_promisc_en(cep->eth,0);
  }
  for (i=0;i<64;i++) {
    hash[i] = (dev->flags & IFF_ALLMULTI) ? 1 : 0;
  }
  for (mc_ptr = dev->mc_list; mc_ptr; mc_ptr = mc_ptr->next) {
    hash[mac_hash(mc_ptr->dmi_addr)] = 1;
  }
  for (i=0;i<64;i++) {
    eth_multihash(cep->eth,i,hash[i]);
  }
}

// Initialize the Altera Ethernet MAC.
static int __init aeth_probe(struct net_device *dev) {
	struct aeth_private *cep;

	SET_MODULE_OWNER(dev);
	cep = (struct aeth_private *)dev->priv;
	if (dev == NULL) {
		/*
		 * Don't allocate the private data here, it is done later
		 * This makes it easier to free the memory when this driver
		 * is used as a module.
		 */
		dev = init_etherdev(0, 0);
		if (dev == NULL)
			return -ENOMEM;
	}

	/* Initialize the device structure. 
	 */
	if (dev->priv == NULL) {
		cep = (struct aeth_private *)kmalloc(sizeof(*cep), GFP_KERNEL);
		dev->priv = cep;
		if (dev->priv == NULL)
			return -ENOMEM;
	}

	__clear_user(cep, sizeof(*cep));

	if (dev->base_addr == 0x0)
		dev->base_addr = (unsigned long)__ioremap(ETH_BASE_ADD, AETH_TOTAL_SIZE, 0);
	else
		dev->base_addr = (unsigned long)__ioremap((unsigned long)dev->base_addr, AETH_TOTAL_SIZE, 0);

	if(request_region(dev->base_addr, AETH_TOTAL_SIZE, cardname) == 0) {
		printk("%s: unable to request region!\n", dev->name);
		return -EAGAIN;
	}
	cep->eth = (unsigned *)(dev->base_addr);
	cep->in_tx = 0;
	// Set default ethernet station address to ep93xx eth addr + 1.
	{
		unsigned char *epmac = (unsigned char *)IO_ADDRESS(MAC_INDAD);
		unsigned long mac = epmac[5] | (epmac[4] << 8) | (epmac[3] << 16);

		mac++;

        	*(unsigned long *)IO_ADDRESS(MAC_AFP) = 0x00; 
		dev->dev_addr[0] = epmac[0];
		dev->dev_addr[1] = epmac[1];
		dev->dev_addr[2] = epmac[2];
		dev->dev_addr[3] = (mac >> 16) & 0xff;
		dev->dev_addr[4] = (mac >> 8) & 0xff;
		dev->dev_addr[5] = mac & 0xff;

		aeth_init(cep->eth,dev->dev_addr);
	}

	//Fill in the fields of the device structure with ethernet values. 
	ether_setup(dev);

	// The Open Ethernet specific entries in the device structure. 
	dev->open = aeth_open;
	dev->hard_start_xmit = aeth_start_xmit;
	dev->stop = aeth_close;
	dev->get_stats = aeth_get_stats;
	dev->set_multicast_list = aeth_set_multicast_list;
	//dev->set_mac_address = aeth_set_mac_add;

	return 0;
}

#ifdef MODULE

static struct net_device dev_aeth = { 
	init: aeth_probe 
};
	
static int io = ETH_BASE_ADD;
static int irq = IRQ_ETH_0;

MODULE_PARM(io, "i");
MODULE_PARM(irq, "i");

MODULE_LICENSE("GPL");

int init_module(void)
{
	int result;

	dev_aeth.base_addr = io;
	dev_aeth.irq       = irq;
	dev_aeth.init      = aeth_probe;

	if ((result = register_netdev(&dev_aeth)) != 0)
		return result;

	return 0;
}

void cleanup_module(void)
{
	free_irq(dev_aeth.irq, &dev_aeth);
	release_region(dev_aeth.base_addr, AETH_TOTAL_SIZE);
	
	if (dev_aeth.priv)
		kfree(dev_aeth.priv);

	unregister_netdev(&dev_aeth);		
}

#endif /* MODULE */

