/*
 *  linux/drivers/serial/tsuart-corestreet.c
 *
 *  TS-UART loader for Technologic Systems TS-CORE-STREET
 *
 *  (c) 2007 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2007-08-31:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"

// fast
// ADRS: 0x62c0 0002 (read only)
// DATA: 0x60c0 0000
// slow
// ADRS: 0x12c0 0002

// address of stat/data is 0x80,0x81
//  
volatile unsigned char *adrs_reg;

static inline unsigned char READ_ADRS(void) {
  return *adrs_reg;
}
static inline void WRITE_ADRS(unsigned char value) {
  *adrs_reg = value;
}
static inline unsigned char READ_DATA(struct tsuart_port *port) {
  return *((volatile unsigned char *)port->u.membase);
}
static inline void WRITE_DATA(struct tsuart_port *port,unsigned char value) {
  *((volatile unsigned char *)port->u.membase) = value;
}


unsigned tsuartGetCST8(struct tsuart_port *port,int offset) {
  unsigned long flags;
  unsigned char adrs;
  unsigned value;

  save_and_cli(flags);                    // disable interrupts, saving previous state
  adrs = READ_ADRS();                     // save current value of address register
  WRITE_ADRS(0x80+offset);                // set address
  value = (unsigned char)READ_DATA(port); // read data register
  WRITE_ADRS(adrs);                       // restore previous value of address register
  restore_flags(flags);                   // restore interrupts to previous state
  return value;
}
void tsuartPutCST8(struct tsuart_port *port,int offset,unsigned value) 
{
  unsigned long flags;
  unsigned char adrs;

  save_and_cli(flags);                     // disable interrupts, saving previous state
  adrs = READ_ADRS();                      // save current value of address register
  WRITE_ADRS(0x80+offset);                 // set address
  WRITE_DATA(port,(unsigned char)value);   // write data register
  WRITE_ADRS(adrs);                        // restore previous value of address register
  restore_flags(flags);                    // restore interrupts to previous state
}

#define BOARD_ID_CORESTREET 0xC083
static int tryInitPort(void) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }

  adrs_reg = ((unsigned char *) __ioremap(0x12c00000, 4096,0)) + 2;
  if (adrs_reg == (unsigned char *)0x12c00002) {
    printk(KERN_ERR "tsuart-corestreet: cannot map adrs memory\n");
    return -ENOMEM;
  }
  // now do magic knock sequence to enable board
  *(adrs_reg-2) = 0xff;
  *(adrs_reg-1) = 0x17;
  *(adrs_reg+1) = 0x07;
  {
    volatile unsigned char *tmp;

    tmp = ((unsigned char *) __ioremap(0x12000000, 4096,0));
    if (tmp == (unsigned char *)0x12000000) {
      printk(KERN_ERR "tsuart-corestreet: cannot map 0x1200_0000\n");
      return -ENOMEM;
    }
    *tmp = 8;
    //__iounmap(tmp);
  }
  {
    volatile unsigned long *tmp;

    tmp = ((unsigned long *) __ioremap(0x80080000, 4096,0))+ 0x07;
    if (tmp == (unsigned long *)0x80080007) {
      printk(KERN_ERR "tsuart-corestreet: cannot map 0x8008_0007\n");
      return -ENOMEM;
    }
    *tmp = 0xffef;
    //__iounmap(tmp);
  }

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetCST8;
  ts->put = tsuartPutCST8;
  ts->mdmctrl = 0;
  ts->debugFlags = portDebug;
  ts->u.mapbase = 0x70c00000;
  ts->u.irq = 32; // 33; // UNKNOWN, FIXME 32,33,40
  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_CORESTREET;

  if (tsuart_register_port(0,ts) != 0) {
    return 0;
  }

  return 1;
}

static int __init tsuart_init(void)
{
  int ports=0;
	
  if (0) {
    printk("TS-UART/CORESTREET did not detect a CORE STREET board\n");
    return -1;
  } else {
    printk("TS-UART/CORESTREET assuming a CORE STREET board\n");
  }
  if (tryInitPort()) {
    ports++;
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;

