/*
 *  linux/drivers/serial/tsuart-dmx.c
 *
 *  TS-UART loader for Technologic Systems DMX TS-7300 bitstream
 *
 *  (c) 2007 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2007-06-29:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include "tsuart.h"

struct uart_driver *dr8=0,*dr9=0;

#define BOARD_ID_DMX 0x0311
#define PORT_STATUS_DMXBRK 64
#define DEBUG_DATA 64

unsigned tsuartGetMEM16_9DMX(struct tsuart_port *port,int offset) {
  unsigned val0 = *((volatile unsigned short *)port->u.membase + 0);
  unsigned val = *((volatile unsigned short *)port->u.membase + offset);

    if (val0 & (1<<11)) {
      port->localStatus |= PORT_STATUS_DMXBRK;
    }
  if (offset == 0) {
    return val & 0x1FF;
  } else {
    if (offset == 1 && port->localStatus & PORT_STATUS_DMXBRK) {
      port->localStatus &= ~PORT_STATUS_DMXBRK;
      val |= (1 << 11);
    }
    return val;
  }
}

static inline int bit_TRE(unsigned short reg_value) {
  return (reg_value & 0x1000) == 0x1000;
}

void tsuartPutMEM16_9DMX(struct tsuart_port *port,int offset,unsigned value) {
  int i=0;

  if (offset == 1 && (port->pstatus & PORT_STATUS_TXMSB) && (value & (1 << 3))) {
    // wait for TRE to tell us the transmitter is idle
    while (!bit_TRE(*((volatile unsigned short *)port->u.membase)) && i < 1000000) i++;
    if (i >= 1000000) printk("Timeout waiting for TRE!\n");
    // send DMX break
    *((volatile unsigned short *)port->u.membase + 0) |= (1<<10);
    udelay(92);
    *((volatile unsigned short *)port->u.membase + 0) &= ~(1<<10);
    udelay(12);
    value &= ~(1 << 3);
  }
  tsuartPutMEM16_9(port,offset,value);
}

/*
  Need to override ts2->get as follows:
    function needs to set a static bit whenever stat bit 11 is set
    needs to transfer bit to data and clear whenever data is read

  Need to send DMX break whenever data is sent with bit set
    92us break + 12us idle

  Bit 11 is the DMX break bit

 */
static int tryInitPort(int i) {
  struct tsuart_port *ts1,*ts2;

  ts1 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts1) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  init_tsuart_port(ts1);
  ts1->u.iotype = SERIAL_IO_MEM;
  ts1->get = tsuartGetMEM16;
  ts1->put = tsuartPutMEM16;
  ts1->debugFlags = portDebug;
  ts1->u.mapbase = 0x72000400+i*4;
  ts1->portStatus = &ts1->pstatus;
  ts1->pstatus = 0;
  ts1->boardId = BOARD_ID_DMX;
  ts1->u.irq = 40;
  ts1->mdmctrl = 0;
  ts1->localStatus |= PORT_STATUS_SHIRQ;

  ts2 = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts2) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  init_tsuart_port(ts2);
  ts2->rxsize = 2;
  ts2->dataMask = 0x9FF;
  ts2->u.iotype = SERIAL_IO_MEM;
  ts2->get = tsuartGetMEM16_9DMX;
  ts2->put = tsuartPutMEM16_9DMX;
  ts2->debugFlags = portDebug;
  ts2->u.mapbase = 0x72000400+i*4;
  ts2->portStatus = &ts1->pstatus;
  ts2->pstatus |= PORT_STATUS_TXMSB;
  ts2->boardId = BOARD_ID_DMX;
  ts2->u.irq = 40;
  ts2->mdmctrl = 0;
  ts2->localStatus |= PORT_STATUS_SHIRQ;

  if (tsuart_register_port(dr8,ts1) == 0) {
    //return 1;
  }
  if (tsuart_register_port(dr9,ts2) == 0) {
    //return 1;
  }
  return 0;
}

#define TTY_NAME8 "ttyT8Sv"
#define CUA_NAME8 "cuaT8Sv"
#define TTY_NAME9 "ttyT9Sv"
#define CUA_NAME9 "cuaT9Sv"

static char *tty = TTY_NAME8,*cua = CUA_NAME8;
static char *tty9 = TTY_NAME9,*cua9 = CUA_NAME9;
MODULE_PARM(tty,"s");
MODULE_PARM(cua,"s");
MODULE_PARM(tty9,"s");
MODULE_PARM(cua9,"s");

static int __init tsuart_init(void)
{
  int i,ports=0;
  volatile unsigned short int *cr;

  if (!tty) {
    tty = TTY_NAME8;
  }
  if (!cua) {
    cua = CUA_NAME8;
  }

  cr = (volatile unsigned short *)((int) __ioremap(0x72000000, 4096,0) + 0x422);
  if (*cr != 0x8bb7) {
    printk("TS-UART/DMX did not find DMX board\n");
    __iounmap((void *)cr);
    return -1;
  } else {
    printk("TS-UART/DMX found DMX board\n");
    __iounmap((void *)cr);
  }

  dr8 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr8) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  dr9 = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!dr9) {
    kfree(dr8);
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
 if (!tsuart_configure_driver(dr8,234,128,tty,cua,4)) {
    kfree(dr8);
    kfree(dr9);
    return 0;
  }
  if (!tsuart_configure_driver(dr9,234,132,tty9,cua9,4)) {
    kfree(dr8);
    kfree(dr9);
    return 0;
  }
  for (i=0;i<2;i++) {
    if (!tryInitPort(i)) {
      ports++;
    }
  }
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
  // This module can be safely unloaded after initialization is complete.
  // We allow module to load only so error conditions can be detected.
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;
