/*
 *  linux/drivers/serial/tsuart-pac4.c
 *
 *  TS-UART loader for Technologic Systems TS-7260
 *
 *  (c) 2006 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2006-05-12:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"


/*
  We define two (OS) ports which share a common hardware port.
  
  We already have a case where two ports share a common hardware port,
  in the case of the TS-7300 8/bit versus 9-bit modes.  In this case,
  each port correctly handles (not) receiving data that is not for it.
  However this is done by ensuring that only one port at a time is
  active.  In our case both boths might be active at the same time.
  One way to achieve this is to keep a static variable in our Get function
  that refuses to actually read the data if the STAT register doesn't say
  its for us.
 */
unsigned tsuartGetMUX8(struct tsuart_port *port,int offset) {
  unsigned val = -1;
  int src;
  /*
    if reading the STAT register, mask out the flow control bit since
    its really telling us whether data is for us or not.  use this bit
    to instead condition data ready, so we know whether or not data is
    for us.
   */
  if (offset == 1) { // DATA
    src = bit_CTS(*(volatile unsigned char *)port->u.membase);
    if (src == port->u.line) {
      return -1;
    }
    
    val = *((volatile unsigned char *)port->u.membase + offset);
  } else if (offset == 0) { // STAT
    val = *((volatile unsigned char *)port->u.membase + offset);
    
    if (bit_CTS(val) == port->u.line) {
      val = val & 0xFD;
    }
    
    val = val | 0x08; // sorry, no constant, CTS wasn't writable
  }
  return val;
}

void tsuartPutMUX8(struct tsuart_port *port,int offset,unsigned value) {
  *((volatile unsigned char *)port->u.membase + offset) = value;
}

struct uart_driver *drc=0,*drd=0;

#define BOARD_ID_PAC4 0x7404
static int tryInitPort(int base,struct uart_driver *dr) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  if (dr != 0) {
    ts->get = tsuartGetMUX8;
    ts->put = tsuartPutMUX8;
  }  else {
    ts->get = tsuartGetMEM8;
    ts->put = tsuartPutMEM8;
  }
  ts->debugFlags = portDebug;
  ts->u.mapbase = /*0 x72000000 + */ base;
  ts->u.irq = 32;
  ts->mdmctrl = 0;
  ts->boardId = BOARD_ID_PAC4;
  if (dr) {
    //    ts->localStatus |= PORT_STATUS_SHIRQ;
  }
  if (tsuart_register_port(dr,ts) != 0) {
    return 0;
  }
  return 1;
}

static int __init tsuart_init(void)
{
  int ports=0;
	
  if (get_ts_model() != TS_MODEL_PAC4) {
    printk("TS-UART/PAC4 did not detect a TS-PAC4 board\n");
    return -1;
  } else {
    printk("TS-UART/PAC4 detected a TS-PAC4 board\n");
  }

  drc = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!drc) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
  drd = kmalloc(sizeof(struct uart_driver),GFP_KERNEL);
  if (!drd) {
    kfree(drc);
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }
 if (!tsuart_configure_driver(drc,234,100,"ttyTSP","cuaTSC",2)) {
    kfree(drc);
    kfree(drd);
    return 0;    
  }
 /*
  if (!tsuart_configure_driver(drd,234,101,"ttyTSD","cuaTSD",1)) {
    kfree(drc);
    kfree(drd);
    return 0;    
  }
 */
  if (tryInitPort(0x12400000,0)) {
    ports++;
  }

  if (tryInitPort(0x12400002,drc)) {
    ports++;
  }

  if (tryInitPort(0x12400002,drc)) {
    ports++;
  }

  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
  // not safe to unload!
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;

