/*
 *  linux/drivers/serial/tsuart7260.c
 *
 *  TS-UART loader for Technologic Systems TS-7260
 *
 *  (c) 2006 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2006-05-12:mos:initial version completed
 *
 */
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
//#include <serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include "tsuart.h"

#define TS_TTL_REG   0x2000000

static inline int get_tsuart_TRI(void) {
  return ((inw(TS7XXX_IO16_BASE + 0x2400000) & 1) == 1);
}

static inline int get_tsuart_ttl(void) {
  unsigned char c;
  c = inb(TS7XXX_IO8_BASE + TS_TTL_REG);
  return c;
}

static inline void put_tsuart_ttl(int flag) {
  if (flag) {
    outb(get_tsuart_ttl() | 64,TS7XXX_IO8_BASE + TS_TTL_REG);
  } else {
    outb(get_tsuart_ttl() & (~64),TS7XXX_IO8_BASE + TS_TTL_REG);
  }
}

//-----------------------------------------------------------------------
//                                                                      /
// /proc Routines                                                       \_
//                                                                      /
//                                                                      \_
//                                                                      /
//                                                                      \_
//-----------------------------------------------------------------------
/*
  The user can set COM3 to TTL levels as follows:
    echo "1" > /proc/driver/tsuart/com3ttl
  The user can set COM3 to RS-232 levels as follows:
    echo "0" > /proc/driver/tsuart/com3ttl
  The user can query the current status of COM3 as follows:
    cat /proc/driver/tsuart/com3ttl
      "0" if COM3 is at RS-232 levels
      "1" if COM3 is at TTL levels

  Note: If this module is unloaded, the serial port is still accessible,
  however the TTL bit will no longer be changeable until reload.
 */
int procfile_read(char *buffer, 
		  char **buffer_location, 
		  off_t offset, 
		  int buffer_length, 
		  int *eof,
		  void *data)
{
  if (offset > 0) {
    *eof = 1;
    return 0;
  }
  MOD_INC_USE_COUNT;
  buffer[0] = (get_tsuart_ttl() & 64) ? '1' : '0';
  *buffer_location = buffer;
  *eof = 1;
  MOD_DEC_USE_COUNT;

  return 1;
}

int procfile_write(struct file *file,
		   const char *buffer,
		   unsigned long count,
		   void *data) {
  char tmp;

  MOD_INC_USE_COUNT;
  if (copy_from_user(&tmp,buffer,1)) {
    return -EFAULT;
  }
  if (tmp == '1') {
    put_tsuart_ttl(1);
  } else if (tmp == '0') {
    put_tsuart_ttl(0);
  }
  MOD_DEC_USE_COUNT;
  return 1;
}

int procfile_init(void) {
  struct proc_dir_entry *dir1,*file;

  dir1 = proc_mkdir("driver/tsuart",NULL);
  if (dir1 == NULL) { return -ENOMEM; }
  
  file = create_proc_entry("driver/tsuart/com3ttl",
			   S_IWUSR |S_IRUSR | S_IRGRP | S_IROTH,
			   NULL);
  if (file == NULL) {
    remove_proc_entry("driver/tsuart", NULL);
    return -ENOMEM;
  }
  file->read_proc = procfile_read;
  file->write_proc = procfile_write;
  file->owner = THIS_MODULE;
  return 0;
}

void procfile_exit(void) {
  remove_proc_entry("driver/tsuart/com3ttl",NULL);
  remove_proc_entry("driver/tsuart", NULL);
}

#define BOARD_ID_7260 0x7260
static int tryInitPort(int base) {
  struct tsuart_port *ts;

  ts = kmalloc(sizeof(struct tsuart_port),GFP_KERNEL);
  if (!ts) {
    printk("Unable to allocate memory for tsuart port\n");
    return 0;
  }

  init_tsuart_port(ts);
  ts->u.iotype = SERIAL_IO_MEM;
  ts->get = tsuartGetMEM8;
  ts->put = tsuartPutMEM8;
  ts->mdmctrl = 0;
  ts->debugFlags = portDebug;
  ts->u.mapbase = /*0 x72000000 + */ base;
  ts->u.irq = 32;
  ts->localStatus |= PORT_STATUS_SHIRQ;
  ts->boardId = BOARD_ID_7260;

  if (tsuart_register_port(0,ts) != 0) {
    return 0;
  }
  return 1;
}

static int __init tsuart_init(void)
{
  int ports=0;
  int ret;
	
  if (get_ts_model() != TS_MODEL_7260) {
    printk("TS-UART/7260 did not detect a TS-7260 board\n");
    return -1;
  } else {
    printk("TS-UART/7260 detected a TS-7260 board\n");
  }
  ret = procfile_init();
  if (ret) {
    return ret;
  }

  cli(); // bug fix for second uart always being on at startup
  if (tryInitPort(0x12400000)) {
    ports++;
  }
  if (get_tsuart_TRI()) {
	  if (tryInitPort(0x13000000)) {
	    ports++;
	  }
	  if (tryInitPort(0x13400000)) {
	    ports++;
	  }
  }
  sti();
  return (ports > 0) ? 0 : -1;
}

static void __exit tsuart_exit(void)
{
  procfile_exit();  
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;

